@@ -53,6 +53,7 @@
 #include <string.h>
 
 #include <glib.h>
+#include <gtk/gtk.h>
 
 #include <clutter/clutter.h>
 
@@ -62,10 +63,11 @@
 #include "st-icon.h"
 #include "st-widget.h"
 #include "st-texture-cache.h"
-#include "st-marshal.h"
 #include "st-clipboard.h"
 #include "st-private.h"
 
+#include "st-widget-accessible.h"
+
 #define HAS_FOCUS(actor) (clutter_actor_get_stage (actor) && clutter_stage_get_key_focus ((ClutterStage *) clutter_actor_get_stage (actor)) == actor)
 
 
@@ -88,9 +90,11 @@ enum
   LAST_SIGNAL
 };
 
-#define ST_ENTRY_GET_PRIVATE(obj)     (G_TYPE_INSTANCE_GET_PRIVATE ((obj), ST_TYPE_ENTRY, StEntryPrivate))
 #define ST_ENTRY_PRIV(x) ((StEntry *) x)->priv
 
+static void         st_entry_check_cursor_blink       (StEntry       *entry);
+static void         st_entry_pend_cursor_blink        (StEntry       *entry);
+static void         st_entry_reset_blink_time         (StEntry       *entry);
 
 struct _StEntryPrivate
 {
@@ -104,11 +108,16 @@ struct _StEntryPrivate
 
   gboolean      hint_visible;
   gboolean      capslock_warning_shown;
+  guint         blink_time;
+  guint         blink_timeout;
+  gboolean      cursor_visible;
 };
 
 static guint entry_signals[LAST_SIGNAL] = { 0, };
 
-G_DEFINE_TYPE (StEntry, st_entry, ST_TYPE_WIDGET);
+G_DEFINE_TYPE_WITH_PRIVATE (StEntry, st_entry, ST_TYPE_WIDGET);
+
+static GType st_entry_accessible_get_type (void) G_GNUC_CONST;
 
 static void
 st_entry_set_property (GObject      *gobject,
@@ -210,6 +219,12 @@ st_entry_dispose (GObject *object)
   StEntryPrivate *priv = entry->priv;
   GdkKeymap *keymap;
 
+  if (priv->blink_timeout)
+    {
+      g_source_remove (priv->blink_timeout);
+      priv->blink_timeout = 0;
+    }
+
   if (priv->entry)
     {
       clutter_actor_destroy (priv->entry);
@@ -239,14 +254,9 @@ st_entry_style_changed (StWidget *self)
   StEntryPrivate *priv = ST_ENTRY_PRIV (self);
   StThemeNode *theme_node;
   ClutterColor color;
-  const PangoFontDescription *font;
-  gchar *font_string;
   gdouble size;
 
   theme_node = st_widget_get_theme_node (self);
- 
-  st_theme_node_get_foreground_color (theme_node, &color);
-  clutter_text_set_color (CLUTTER_TEXT (priv->entry), &color);
 
   if (st_theme_node_lookup_length (theme_node, "caret-size", TRUE, &size))
     clutter_text_set_cursor_size (CLUTTER_TEXT (priv->entry), (int)(.5 + size));
@@ -260,10 +270,7 @@ st_entry_style_changed (StWidget *self)
   if (st_theme_node_lookup_color (theme_node, "selected-color", TRUE, &color))
     clutter_text_set_selected_text_color (CLUTTER_TEXT (priv->entry), &color);
 
-  font = st_theme_node_get_font (theme_node);
-  font_string = pango_font_description_to_string (font);
-  clutter_text_set_font_name (CLUTTER_TEXT (priv->entry), font_string);
-  g_free (font_string);
+  _st_set_text_from_style ((ClutterText *)priv->entry, theme_node);
 
   ST_WIDGET_CLASS (st_entry_parent_class)->style_changed (self);
 }
@@ -383,13 +390,11 @@ st_entry_allocate (ClutterActor          *actor,
 {
   StEntryPrivate *priv = ST_ENTRY_PRIV (actor);
   StThemeNode *theme_node = st_widget_get_theme_node (ST_WIDGET (actor));
-  ClutterActorClass *parent_class;
   ClutterActorBox content_box, child_box, icon_box;
   gfloat icon_w, icon_h;
   gfloat entry_h, min_h, pref_h, avail_h;
 
-  parent_class = CLUTTER_ACTOR_CLASS (st_entry_parent_class);
-  parent_class->allocate (actor, box, flags);
+  clutter_actor_set_allocation (actor, box, flags);
 
   st_theme_node_get_content_box (theme_node, box, &content_box);
 
@@ -416,7 +421,7 @@ st_entry_allocate (ClutterActor          *actor,
                               flags);
 
       /* reduce the size for the entry */
-      child_box.x1 += icon_w + priv->spacing;
+      child_box.x1 = MIN (child_box.x2, child_box.x1 + icon_w + priv->spacing);
     }
 
   if (priv->secondary_icon)
@@ -437,7 +442,7 @@ st_entry_allocate (ClutterActor          *actor,
                               flags);
 
       /* reduce the size for the entry */
-      child_box.x2 -= icon_w - priv->spacing;
+       child_box.x2 = MAX (child_box.x1, child_box.x2 - icon_w - priv->spacing);
     }
 
   clutter_actor_get_preferred_height (priv->entry, child_box.x2 - child_box.x1,
@@ -451,6 +456,181 @@ st_entry_allocate (ClutterActor          *actor,
   clutter_actor_allocate (priv->entry, &child_box, flags);
 }
 
+#define CURSOR_ON_MULTIPLIER 2
+#define CURSOR_OFF_MULTIPLIER 1
+#define CURSOR_PEND_MULTIPLIER 3
+#define CURSOR_DIVIDER 3
+
+static gboolean
+cursor_blinks (StEntry *entry)
+{
+  StEntryPrivate *priv = entry->priv;
+
+  if (clutter_actor_has_key_focus (CLUTTER_ACTOR (priv->entry)) &&
+      clutter_text_get_editable (CLUTTER_TEXT (priv->entry)) &&
+      clutter_text_get_selection_bound (CLUTTER_TEXT (priv->entry)) == clutter_text_get_cursor_position (CLUTTER_TEXT (priv->entry)))
+    {
+      gboolean blink;
+      g_object_get (gtk_settings_get_default (), "gtk-cursor-blink", &blink, NULL);
+
+      return blink;
+    }
+  else
+    return FALSE;
+}
+
+static gint
+get_cursor_time (StEntry *entry)
+{
+  gint time;
+  g_object_get (gtk_settings_get_default (), "gtk-cursor-blink-time", &time, NULL);
+
+  return time;
+}
+
+static gint
+get_cursor_blink_timeout (StEntry *entry)
+{
+  gint timeout;
+  g_object_get (gtk_settings_get_default (), "gtk-cursor-blink-timeout", &timeout, NULL);
+
+  return timeout;
+}
+
+static void
+show_cursor (StEntry *entry)
+{
+  StEntryPrivate *priv = entry->priv;
+
+  if (!priv->cursor_visible)
+    {
+      priv->cursor_visible = TRUE;
+      clutter_text_set_cursor_visible (CLUTTER_TEXT (priv->entry), TRUE);
+    }
+}
+
+static void
+hide_cursor (StEntry *entry)
+{
+  StEntryPrivate *priv = entry->priv;
+
+  if (priv->cursor_visible)
+    {
+      priv->cursor_visible = FALSE;
+      clutter_text_set_cursor_visible (CLUTTER_TEXT (priv->entry), FALSE);
+    }
+}
+
+/*
+ * Blink!
+ */
+static gint
+blink_cb (gpointer data)
+{
+  StEntry *entry;
+  StEntryPrivate *priv;
+  guint blink_timeout;
+
+  entry = ST_ENTRY (data);
+  priv = entry->priv;
+
+  if (!clutter_actor_has_key_focus (priv->entry))
+    {
+      g_warning ("StEntry - did not receive key-focus-out event. If you\n"
+         "connect a handler to this signal, it must return\n"
+         "FALSE so the StEntry gets the event as well");
+
+      st_entry_check_cursor_blink (entry);
+
+      return FALSE;
+    }
+
+  if (clutter_text_get_selection_bound (CLUTTER_TEXT (priv->entry)) != clutter_text_get_cursor_position (CLUTTER_TEXT (priv->entry)))
+    {
+      st_entry_check_cursor_blink (entry);
+      return FALSE;
+    }
+
+  blink_timeout = get_cursor_blink_timeout (entry);
+  if (priv->blink_time > 1000 * blink_timeout &&
+      blink_timeout < G_MAXINT/1000)
+    {
+      /* we've blinked enough without the user doing anything, stop blinking */
+      show_cursor (entry);
+      priv->blink_timeout = 0;
+    }
+  else if (priv->cursor_visible)
+    {
+      hide_cursor (entry);
+      priv->blink_timeout = clutter_threads_add_timeout (get_cursor_time (entry) * CURSOR_OFF_MULTIPLIER / CURSOR_DIVIDER,
+                        blink_cb,
+                        entry);
+    }
+  else
+    {
+      show_cursor (entry);
+      priv->blink_time += get_cursor_time (entry);
+      priv->blink_timeout = clutter_threads_add_timeout (get_cursor_time (entry) * CURSOR_ON_MULTIPLIER / CURSOR_DIVIDER,
+                        blink_cb,
+                        entry);
+    }
+
+  /* Remove ourselves */
+  return FALSE;
+}
+
+static void
+st_entry_check_cursor_blink (StEntry *entry)
+{
+  StEntryPrivate *priv = entry->priv;
+
+  if (cursor_blinks (entry))
+    {
+      if (!priv->blink_timeout)
+    {
+      show_cursor (entry);
+      priv->blink_timeout = clutter_threads_add_timeout (get_cursor_time (entry) * CURSOR_ON_MULTIPLIER / CURSOR_DIVIDER,
+                        blink_cb,
+                        entry);
+    }
+    }
+  else
+    {
+      if (priv->blink_timeout)
+        {
+          g_source_remove (priv->blink_timeout);
+          priv->blink_timeout = 0;
+        }
+
+      show_cursor (entry);
+    }
+}
+
+static void
+st_entry_pend_cursor_blink (StEntry *entry)
+{
+  StEntryPrivate *priv = entry->priv;
+
+  if (cursor_blinks (entry))
+    {
+      if (priv->blink_timeout != 0)
+    g_source_remove (priv->blink_timeout);
+
+      priv->blink_timeout = clutter_threads_add_timeout (get_cursor_time (entry) * CURSOR_PEND_MULTIPLIER / CURSOR_DIVIDER,
+                                                     blink_cb,
+                                                     entry);
+      show_cursor (entry);
+    }
+}
+
+static void
+st_entry_reset_blink_time (StEntry *entry)
+{
+  StEntryPrivate *priv = entry->priv;
+
+  priv->blink_time = 0;
+}
+
 static void
 clutter_text_focus_in_cb (ClutterText  *text,
                           ClutterActor *actor)
@@ -474,7 +654,9 @@ clutter_text_focus_in_cb (ClutterText  *text,
 
   st_widget_remove_style_pseudo_class (ST_WIDGET (actor), "indeterminate");
   st_widget_add_style_pseudo_class (ST_WIDGET (actor), "focus");
-  clutter_text_set_cursor_visible (text, TRUE);
+
+  st_entry_reset_blink_time (entry);
+  st_entry_check_cursor_blink (entry);
 }
 
 static void
@@ -495,7 +677,7 @@ clutter_text_focus_out_cb (ClutterText  *text,
       clutter_text_set_text (text, priv->hint);
       st_widget_add_style_pseudo_class (ST_WIDGET (actor), "indeterminate");
     }
-  clutter_text_set_cursor_visible (text, FALSE);
+  st_entry_check_cursor_blink (entry);
   remove_capslock_feedback (entry);
 
   keymap = gdk_keymap_get_for_display (gdk_display_get_default ());
@@ -514,38 +696,21 @@ clutter_text_password_char_cb (GObject    *object,
 }
 
 static void
-st_entry_paint (ClutterActor *actor)
+clutter_text_selection_bound_cb (GObject    *object,
+                                 GParamSpec *pspec,
+                                 gpointer    user_data)
 {
-  StEntryPrivate *priv = ST_ENTRY_PRIV (actor);
-  ClutterActorClass *parent_class;
+  StEntry *entry = ST_ENTRY (user_data);
 
-  parent_class = CLUTTER_ACTOR_CLASS (st_entry_parent_class);
-  parent_class->paint (actor);
-
-  clutter_actor_paint (priv->entry);
-
-  if (priv->primary_icon)
-    clutter_actor_paint (priv->primary_icon);
-
-  if (priv->secondary_icon)
-    clutter_actor_paint (priv->secondary_icon);
+  st_entry_reset_blink_time (entry);
+  st_entry_pend_cursor_blink (entry);
 }
 
 static void
-st_entry_pick (ClutterActor       *actor,
-               const ClutterColor *c)
+clutter_text_cursor_changed (ClutterText *text, ClutterActor *actor)
 {
-  StEntryPrivate *priv = ST_ENTRY_PRIV (actor);
-
-  CLUTTER_ACTOR_CLASS (st_entry_parent_class)->pick (actor, c);
-
-  clutter_actor_paint (priv->entry);
-
-  if (priv->primary_icon)
-    clutter_actor_paint (priv->primary_icon);
-
-  if (priv->secondary_icon)
-    clutter_actor_paint (priv->secondary_icon);
+  st_entry_reset_blink_time (ST_ENTRY (actor));
+  st_entry_pend_cursor_blink (ST_ENTRY (actor));
 }
 
 static void
@@ -567,32 +732,73 @@ st_entry_clipboard_callback (StClipboard *clipboard,
   clutter_text_insert_text (ctext, text, cursor_pos);
 }
 
+static gboolean
+clutter_text_button_press_event (ClutterActor       *actor,
+                                 ClutterButtonEvent *event,
+                                 gpointer            user_data)
+{
+  StEntryPrivate *priv = ST_ENTRY_PRIV (user_data);
+  GtkSettings *settings = gtk_settings_get_default ();
+  gboolean primary_paste_enabled;
+
+  g_object_get (settings,
+                "gtk-enable-primary-paste", &primary_paste_enabled,
+                NULL);
+
+  if (primary_paste_enabled && event->button == 2
+      && clutter_text_get_editable (CLUTTER_TEXT (priv->entry)))
+    {
+      StClipboard *clipboard;
+
+      clipboard = st_clipboard_get_default ();
+
+      /* By the time the clipboard callback is called,
+       * the rest of the signal handlers will have
+       * run, making the text cursor to be in the correct
+       * place.
+       */
+      st_clipboard_get_text (clipboard,
+                             ST_CLIPBOARD_TYPE_PRIMARY,
+                             st_entry_clipboard_callback,
+                             user_data);
+    }
+
+  return FALSE;
+}
+
+
 static gboolean
 st_entry_key_press_event (ClutterActor    *actor,
                           ClutterKeyEvent *event)
 {
   StEntryPrivate *priv = ST_ENTRY_PRIV (actor);
 
+  st_entry_reset_blink_time (ST_ENTRY (actor));
+  st_entry_pend_cursor_blink (ST_ENTRY (actor));
+
   /* This is expected to handle events that were emitted for the inner
      ClutterText. They only reach this function if the ClutterText
      didn't handle them */
 
   /* paste */
   if ((event->modifier_state & CLUTTER_CONTROL_MASK)
-      && event->keyval == CLUTTER_v)
+      && (event->keyval == CLUTTER_KEY_v || event->keyval == CLUTTER_KEY_V))
     {
       StClipboard *clipboard;
 
       clipboard = st_clipboard_get_default ();
 
-      st_clipboard_get_text (clipboard, st_entry_clipboard_callback, actor);
+      st_clipboard_get_text (clipboard,
+                             ST_CLIPBOARD_TYPE_CLIPBOARD,
+                             st_entry_clipboard_callback,
+                             actor);
 
       return TRUE;
     }
 
   /* copy */
   if ((event->modifier_state & CLUTTER_CONTROL_MASK)
-      && event->keyval == CLUTTER_c)
+      && (event->keyval == CLUTTER_KEY_c || event->keyval == CLUTTER_KEY_C))
     {
       StClipboard *clipboard;
       gchar *text;
@@ -602,7 +808,9 @@ st_entry_key_press_event (ClutterActor    *actor,
       text = clutter_text_get_selection ((ClutterText*) priv->entry);
 
       if (text && strlen (text))
-        st_clipboard_set_text (clipboard, text);
+        st_clipboard_set_text (clipboard,
+                               ST_CLIPBOARD_TYPE_CLIPBOARD,
+                               text);
 
       return TRUE;
     }
@@ -610,7 +818,7 @@ st_entry_key_press_event (ClutterActor    *actor,
 
   /* cut */
   if ((event->modifier_state & CLUTTER_CONTROL_MASK)
-      && event->keyval == CLUTTER_x)
+      && (event->keyval == CLUTTER_KEY_x || event->keyval == CLUTTER_KEY_X))
     {
       StClipboard *clipboard;
       gchar *text;
@@ -621,7 +829,7 @@ st_entry_key_press_event (ClutterActor    *actor,
 
       if (text && strlen (text))
         {
-          st_clipboard_set_text (clipboard, text);
+          st_clipboard_set_text (clipboard,ST_CLIPBOARD_TYPE_CLIPBOARD, text);
 
           /* now delete the text */
           clutter_text_delete_selection ((ClutterText *) priv->entry);
@@ -651,8 +859,6 @@ st_entry_class_init (StEntryClass *klass)
   StWidgetClass *widget_class = ST_WIDGET_CLASS (klass);
   GParamSpec *pspec;
 
-  g_type_class_add_private (klass, sizeof (StEntryPrivate));
-
   gobject_class->set_property = st_entry_set_property;
   gobject_class->get_property = st_entry_get_property;
   gobject_class->finalize = st_entry_finalize;
@@ -661,20 +867,19 @@ st_entry_class_init (StEntryClass *klass)
   actor_class->get_preferred_width = st_entry_get_preferred_width;
   actor_class->get_preferred_height = st_entry_get_preferred_height;
   actor_class->allocate = st_entry_allocate;
-  actor_class->paint = st_entry_paint;
-  actor_class->pick = st_entry_pick;
 
   actor_class->key_press_event = st_entry_key_press_event;
   actor_class->key_focus_in = st_entry_key_focus_in;
 
   widget_class->style_changed = st_entry_style_changed;
   widget_class->navigate_focus = st_entry_navigate_focus;
+  widget_class->get_accessible_type = st_entry_accessible_get_type;
 
   pspec = g_param_spec_object ("clutter-text",
-			       "Clutter Text",
-			       "Internal ClutterText actor",
-			       CLUTTER_TYPE_TEXT,
-			       G_PARAM_READABLE);
+             "Clutter Text",
+             "Internal ClutterText actor",
+             CLUTTER_TYPE_TEXT,
+             G_PARAM_READABLE);
   g_object_class_install_property (gobject_class, PROP_CLUTTER_TEXT, pspec);
 
   pspec = g_param_spec_string ("hint-text",
@@ -701,8 +906,7 @@ st_entry_class_init (StEntryClass *klass)
                   G_TYPE_FROM_CLASS (klass),
                   G_SIGNAL_RUN_LAST,
                   G_STRUCT_OFFSET (StEntryClass, primary_icon_clicked),
-                  NULL, NULL,
-                  _st_marshal_VOID__VOID,
+                  NULL, NULL, NULL,
                   G_TYPE_NONE, 0);
   /**
    * StEntry::secondary-icon-clicked:
@@ -714,8 +918,7 @@ st_entry_class_init (StEntryClass *klass)
                   G_TYPE_FROM_CLASS (klass),
                   G_SIGNAL_RUN_LAST,
                   G_STRUCT_OFFSET (StEntryClass, secondary_icon_clicked),
-                  NULL, NULL,
-                  _st_marshal_VOID__VOID,
+                  NULL, NULL, NULL,
                   G_TYPE_NONE, 0);
 }
 
@@ -724,7 +927,7 @@ st_entry_init (StEntry *entry)
 {
   StEntryPrivate *priv;
 
-  priv = entry->priv = ST_ENTRY_GET_PRIVATE (entry);
+  priv = entry->priv = st_entry_get_instance_private (entry);
 
   priv->entry = g_object_new (ST_TYPE_IM_TEXT,
                               "line-alignment", PANGO_ALIGN_LEFT,
@@ -742,12 +945,24 @@ st_entry_init (StEntry *entry)
   g_signal_connect (priv->entry, "notify::password-char",
                     G_CALLBACK (clutter_text_password_char_cb), entry);
 
+  g_signal_connect (priv->entry, "button-press-event",
+                    G_CALLBACK (clutter_text_button_press_event), entry);
+
+  g_signal_connect (priv->entry, "notify::selection-bound",
+                    G_CALLBACK (clutter_text_selection_bound_cb), entry);
+
+  g_signal_connect (priv->entry, "cursor-changed",
+                    G_CALLBACK (clutter_text_cursor_changed), entry);
+
   priv->spacing = 6.0f;
 
-  clutter_actor_set_parent (priv->entry, CLUTTER_ACTOR (entry));
+  clutter_actor_add_child (CLUTTER_ACTOR (entry), priv->entry);
   clutter_actor_set_reactive ((ClutterActor *) entry, TRUE);
 
   /* set cursor hidden until we receive focus */
+  priv->blink_timeout = 0;
+  priv->blink_time = 0;
+  priv->cursor_visible = FALSE;
   clutter_text_set_cursor_visible ((ClutterText *) priv->entry, FALSE);
 }
 
@@ -921,7 +1136,7 @@ _st_entry_set_icon (StEntry       *entry,
       g_signal_handlers_disconnect_by_func (*icon,
                                             _st_entry_icon_press_cb,
                                             entry);
-      clutter_actor_unparent (*icon);
+      clutter_actor_remove_child (CLUTTER_ACTOR (entry), *icon);
       *icon = NULL;
     }
 
@@ -930,7 +1145,7 @@ _st_entry_set_icon (StEntry       *entry,
       *icon = g_object_ref (new_icon);
 
       clutter_actor_set_reactive (*icon, TRUE);
-      clutter_actor_set_parent (*icon, CLUTTER_ACTOR (entry));
+      clutter_actor_add_child (CLUTTER_ACTOR (entry), *icon);
       g_signal_connect (*icon, "button-release-event",
                         G_CALLBACK (_st_entry_icon_press_cb), entry);
     }
@@ -948,10 +1163,17 @@ _st_entry_set_icon_from_file (StEntry       *entry,
   if (filename)
     {
       StTextureCache *cache;
+      GFile *file;
+      gchar *uri;
 
       cache = st_texture_cache_get_default ();
+      file = g_file_new_for_path (filename);
+      uri = g_file_get_uri (file);
 
-      new_icon = (ClutterActor*) st_texture_cache_load_file_simple (cache, filename);
+      g_object_unref (file);
+
+      new_icon = (ClutterActor*) st_texture_cache_load_uri_async (cache, uri, -1, -1);
+      g_free (uri);
     }
 
   _st_entry_set_icon  (entry, icon, new_icon);
@@ -1039,3 +1261,97 @@ st_entry_set_secondary_icon_from_file (StEntry     *entry,
 
 }
 
+/******************************************************************************/
+/*************************** ACCESSIBILITY SUPPORT ****************************/
+/******************************************************************************/
+
+#define ST_TYPE_ENTRY_ACCESSIBLE         (st_entry_accessible_get_type ())
+#define ST_ENTRY_ACCESSIBLE(o)           (G_TYPE_CHECK_INSTANCE_CAST ((o), ST_TYPE_ENTRY_ACCESSIBLE, StEntryAccessible))
+#define ST_IS_ENTRY_ACCESSIBLE(o)        (G_TYPE_CHECK_INSTANCE_TYPE ((o), ST_TYPE_ENTRY_ACCESSIBLE))
+#define ST_ENTRY_ACCESSIBLE_CLASS(c)     (G_TYPE_CHECK_CLASS_CAST ((c),    ST_TYPE_ENTRY_ACCESSIBLE, StEntryAccessibleClass))
+#define ST_IS_ENTRY_ACCESSIBLE_CLASS(c)  (G_TYPE_CHECK_CLASS_TYPE ((c),    ST_TYPE_ENTRY_ACCESSIBLE))
+#define ST_ENTRY_ACCESSIBLE_GET_CLASS(o) (G_TYPE_INSTANCE_GET_CLASS ((o),  ST_TYPE_ENTRY_ACCESSIBLE, StEntryAccessibleClass))
+
+typedef struct _StEntryAccessible  StEntryAccessible;
+typedef struct _StEntryAccessibleClass  StEntryAccessibleClass;
+
+struct _StEntryAccessible
+{
+  StWidgetAccessible parent;
+};
+
+struct _StEntryAccessibleClass
+{
+  StWidgetAccessibleClass parent_class;
+};
+
+G_DEFINE_TYPE (StEntryAccessible, st_entry_accessible, ST_TYPE_WIDGET_ACCESSIBLE)
+
+static void
+st_entry_accessible_init (StEntryAccessible *self)
+{
+  /* initialization done on AtkObject->initialize */
+}
+
+static void
+st_entry_accessible_initialize (AtkObject *obj,
+                                gpointer   data)
+{
+  ATK_OBJECT_CLASS (st_entry_accessible_parent_class)->initialize (obj, data);
+
+  /* StEntry is behaving as a StImText container */
+  atk_object_set_role (obj, ATK_ROLE_PANEL);
+}
+
+static gint
+st_entry_accessible_get_n_children (AtkObject *obj)
+{
+  StEntry *entry = NULL;
+
+  g_return_val_if_fail (ST_IS_ENTRY_ACCESSIBLE (obj), 0);
+
+  entry = ST_ENTRY (atk_gobject_accessible_get_object (ATK_GOBJECT_ACCESSIBLE (obj)));
+
+  if (entry == NULL)
+    return 0;
+
+  if (entry->priv->entry == NULL)
+    return 0;
+  else
+    return 1;
+}
+
+static AtkObject*
+st_entry_accessible_ref_child (AtkObject *obj,
+                               gint       i)
+{
+  StEntry *entry = NULL;
+  AtkObject *result = NULL;
+
+  g_return_val_if_fail (ST_IS_ENTRY_ACCESSIBLE (obj), NULL);
+  g_return_val_if_fail (i == 0, NULL);
+
+  entry = ST_ENTRY (atk_gobject_accessible_get_object (ATK_GOBJECT_ACCESSIBLE (obj)));
+
+  if (entry == NULL)
+    return NULL;
+
+  if (entry->priv->entry == NULL)
+    return NULL;
+
+  result = clutter_actor_get_accessible (entry->priv->entry);
+  g_object_ref (result);
+
+  return result;
+}
+
+
+static void
+st_entry_accessible_class_init (StEntryAccessibleClass *klass)
+{
+  AtkObjectClass *atk_class = ATK_OBJECT_CLASS (klass);
+
+  atk_class->initialize = st_entry_accessible_initialize;
+  atk_class->get_n_children = st_entry_accessible_get_n_children;
+  atk_class->ref_child= st_entry_accessible_ref_child;
+}
